# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("dhcpsapi.dll")
prototypes = \
    {
        #
        'DhcpAddFilterV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_FILTER_ADD_INFO", SimStruct), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "AddFilterInfo", "ForceFlag"]),
        #
        'DhcpDeleteFilterV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_ADDR_PATTERN", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "DeleteFilterInfo"]),
        #
        'DhcpSetFilterV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_FILTER_GLOBAL_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "GlobalFilterInfo"]),
        #
        'DhcpGetFilterV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_FILTER_GLOBAL_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "GlobalFilterInfo"]),
        #
        'DhcpEnumFilterV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_ADDR_PATTERN", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DHCP_FILTER_LIST_TYPE"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_FILTER_ENUM_INFO", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ResumeHandle", "PreferredMaximum", "ListType", "EnumFilterInfo", "ElementsRead", "ElementsTotal"]),
        #
        'DhcpCreateSubnet': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SubnetInfo"]),
        #
        'DhcpSetSubnetInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SubnetInfo"]),
        #
        'DhcpGetSubnetInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SUBNET_INFO", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SubnetInfo"]),
        #
        'DhcpEnumSubnets': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_IP_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ResumeHandle", "PreferredMaximum", "EnumInfo", "ElementsRead", "ElementsTotal"]),
        #
        'DhcpAddSubnetElement': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_DATA", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "AddElementInfo"]),
        #
        'DhcpEnumSubnetElements': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DHCP_SUBNET_ELEMENT_TYPE"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_INFO_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "EnumElementType", "ResumeHandle", "PreferredMaximum", "EnumElementInfo", "ElementsRead", "ElementsTotal"]),
        #
        'DhcpRemoveSubnetElement': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_DATA", SimStruct), offset=0), SimTypeInt(signed=False, label="DHCP_FORCE_FLAG")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "RemoveElementInfo", "ForceFlag"]),
        #
        'DhcpDeleteSubnet': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DHCP_FORCE_FLAG")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ForceFlag"]),
        #
        'DhcpCreateOption': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_OPTION", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "OptionID", "OptionInfo"]),
        #
        'DhcpSetOptionInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_OPTION", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "OptionID", "OptionInfo"]),
        #
        'DhcpGetOptionInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "OptionID", "OptionInfo"]),
        #
        'DhcpEnumOptions': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ResumeHandle", "PreferredMaximum", "Options", "OptionsRead", "OptionsTotal"]),
        #
        'DhcpRemoveOption': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "OptionID"]),
        #
        'DhcpSetOptionValue': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_DATA", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "OptionID", "ScopeInfo", "OptionValue"]),
        #
        'DhcpSetOptionValues': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE_ARRAY", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ScopeInfo", "OptionValues"]),
        #
        'DhcpGetOptionValue': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "OptionID", "ScopeInfo", "OptionValue"]),
        #
        'DhcpEnumOptionValues': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ScopeInfo", "ResumeHandle", "PreferredMaximum", "OptionValues", "OptionsRead", "OptionsTotal"]),
        #
        'DhcpRemoveOptionValue': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "OptionID", "ScopeInfo"]),
        #
        'DhcpCreateClientInfoVQ': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_VQ", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpSetClientInfoVQ': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_VQ", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpGetClientInfoVQ': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_SEARCH_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_VQ", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SearchInfo", "ClientInfo"]),
        #
        'DhcpEnumSubnetClientsVQ': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_ARRAY_VQ", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ResumeHandle", "PreferredMaximum", "ClientInfo", "ClientsRead", "ClientsTotal"]),
        #
        'DhcpEnumSubnetClientsFilterStatusInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_FILTER_STATUS_INFO_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ResumeHandle", "PreferredMaximum", "ClientInfo", "ClientsRead", "ClientsTotal"]),
        #
        'DhcpCreateClientInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpSetClientInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpGetClientInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_SEARCH_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SearchInfo", "ClientInfo"]),
        #
        'DhcpDeleteClientInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_SEARCH_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpEnumSubnetClients': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ResumeHandle", "PreferredMaximum", "ClientInfo", "ClientsRead", "ClientsTotal"]),
        #
        'DhcpGetClientOptions': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_LIST", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientIpAddress", "ClientSubnetMask", "ClientOptions"]),
        #
        'DhcpGetMibInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_MIB_INFO", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "MibInfo"]),
        #
        'DhcpServerSetConfig': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SERVER_CONFIG_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "FieldsToSet", "ConfigInfo"]),
        #
        'DhcpServerGetConfig': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SERVER_CONFIG_INFO", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ConfigInfo"]),
        #
        'DhcpScanDatabase': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SCAN_LIST", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "FixFlag", "ScanList"]),
        #
        'DhcpRpcFreeMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["BufferPointer"]),
        #
        'DhcpGetVersion': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "MajorVersion", "MinorVersion"]),
        #
        'DhcpAddSubnetElementV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_DATA_V4", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "AddElementInfo"]),
        #
        'DhcpEnumSubnetElementsV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DHCP_SUBNET_ELEMENT_TYPE"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_INFO_ARRAY_V4", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "EnumElementType", "ResumeHandle", "PreferredMaximum", "EnumElementInfo", "ElementsRead", "ElementsTotal"]),
        #
        'DhcpRemoveSubnetElementV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_DATA_V4", SimStruct), offset=0), SimTypeInt(signed=False, label="DHCP_FORCE_FLAG")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "RemoveElementInfo", "ForceFlag"]),
        #
        'DhcpCreateClientInfoV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_V4", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpSetClientInfoV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_V4", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpGetClientInfoV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_SEARCH_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_V4", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SearchInfo", "ClientInfo"]),
        #
        'DhcpEnumSubnetClientsV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_ARRAY_V4", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ResumeHandle", "PreferredMaximum", "ClientInfo", "ClientsRead", "ClientsTotal"]),
        #
        'DhcpServerSetConfigV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SERVER_CONFIG_INFO_V4", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "FieldsToSet", "ConfigInfo"]),
        #
        'DhcpServerGetConfigV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SERVER_CONFIG_INFO_V4", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ConfigInfo"]),
        #
        'DhcpSetSuperScopeV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SuperScopeName", "ChangeExisting"]),
        #
        'DhcpDeleteSuperScopeV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SuperScopeName"]),
        #
        'DhcpGetSuperScopeInfoV4': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SUPER_SCOPE_TABLE", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SuperScopeTable"]),
        #
        'DhcpEnumSubnetClientsV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_ARRAY_V5", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ResumeHandle", "PreferredMaximum", "ClientInfo", "ClientsRead", "ClientsTotal"]),
        #
        'DhcpCreateOptionV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionId", "ClassName", "VendorName", "OptionInfo"]),
        #
        'DhcpSetOptionInfoV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName", "OptionInfo"]),
        #
        'DhcpGetOptionInfoV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName", "OptionInfo"]),
        #
        'DhcpEnumOptionsV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "ClassName", "VendorName", "ResumeHandle", "PreferredMaximum", "Options", "OptionsRead", "OptionsTotal"]),
        #
        'DhcpRemoveOptionV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName"]),
        #
        'DhcpSetOptionValueV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_DATA", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionId", "ClassName", "VendorName", "ScopeInfo", "OptionValue"]),
        #
        'DhcpSetOptionValuesV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE_ARRAY", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "ClassName", "VendorName", "ScopeInfo", "OptionValues"]),
        #
        'DhcpGetOptionValueV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName", "ScopeInfo", "OptionValue"]),
        #
        'DhcpGetOptionValueV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO6", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName", "ScopeInfo", "OptionValue"]),
        #
        'DhcpEnumOptionValuesV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "ClassName", "VendorName", "ScopeInfo", "ResumeHandle", "PreferredMaximum", "OptionValues", "OptionsRead", "OptionsTotal"]),
        #
        'DhcpRemoveOptionValueV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName", "ScopeInfo"]),
        #
        'DhcpCreateClass': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_CLASS_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ReservedMustBeZero", "ClassInfo"]),
        #
        'DhcpModifyClass': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_CLASS_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ReservedMustBeZero", "ClassInfo"]),
        #
        'DhcpDeleteClass': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ReservedMustBeZero", "ClassName"]),
        #
        'DhcpGetClassInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_CLASS_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLASS_INFO", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ReservedMustBeZero", "PartialClassInfo", "FilledClassInfo"]),
        #
        'DhcpEnumClasses': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLASS_INFO_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ReservedMustBeZero", "ResumeHandle", "PreferredMaximum", "ClassInfoArray", "nRead", "nTotal"]),
        #
        'DhcpGetAllOptions': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_ALL_OPTIONS", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionStruct"]),
        #
        'DhcpGetAllOptionsV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_ALL_OPTIONS", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionStruct"]),
        #
        'DhcpGetAllOptionValues': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_ALL_OPTION_VALUES", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "ScopeInfo", "Values"]),
        #
        'DhcpGetAllOptionValuesV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO6", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_ALL_OPTION_VALUES", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "ScopeInfo", "Values"]),
        #
        'DhcpEnumServers': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCPDS_SERVERS", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Flags", "IdInfo", "Servers", "CallbackFn", "CallbackData"]),
        #
        'DhcpAddServer': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeRef("DHCPDS_SERVER", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Flags", "IdInfo", "NewServer", "CallbackFn", "CallbackData"]),
        #
        'DhcpDeleteServer': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeRef("DHCPDS_SERVER", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Flags", "IdInfo", "NewServer", "CallbackFn", "CallbackData"]),
        #
        'DhcpGetServerBindingInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_BIND_ELEMENT_ARRAY", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "BindElementsInfo"]),
        #
        'DhcpSetServerBindingInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_BIND_ELEMENT_ARRAY", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "BindElementInfo"]),
        #
        'DhcpAddSubnetElementV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_DATA_V5", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "AddElementInfo"]),
        #
        'DhcpEnumSubnetElementsV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DHCP_SUBNET_ELEMENT_TYPE"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_INFO_ARRAY_V5", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "EnumElementType", "ResumeHandle", "PreferredMaximum", "EnumElementInfo", "ElementsRead", "ElementsTotal"]),
        #
        'DhcpRemoveSubnetElementV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_DATA_V5", SimStruct), offset=0), SimTypeInt(signed=False, label="DHCP_FORCE_FLAG")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "RemoveElementInfo", "ForceFlag"]),
        #
        'DhcpV4EnumSubnetReservations': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_RESERVATION_INFO_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ResumeHandle", "PreferredMaximum", "EnumElementInfo", "ElementsRead", "ElementsTotal"]),
        #
        'DhcpCreateOptionV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionId", "ClassName", "VendorName", "OptionInfo"]),
        #
        'DhcpRemoveOptionV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName"]),
        #
        'DhcpEnumOptionsV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "ClassName", "VendorName", "ResumeHandle", "PreferredMaximum", "Options", "OptionsRead", "OptionsTotal"]),
        #
        'DhcpRemoveOptionValueV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName", "ScopeInfo"]),
        #
        'DhcpGetOptionInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName", "OptionInfo"]),
        #
        'DhcpSetOptionInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "ClassName", "VendorName", "OptionInfo"]),
        #
        'DhcpSetOptionValueV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO6", SimStruct), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_DATA", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionId", "ClassName", "VendorName", "ScopeInfo", "OptionValue"]),
        #
        'DhcpGetSubnetInfoVQ': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SUBNET_INFO_VQ", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SubnetInfo"]),
        #
        'DhcpCreateSubnetVQ': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_INFO_VQ", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SubnetInfo"]),
        #
        'DhcpSetSubnetInfoVQ': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SUBNET_INFO_VQ", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SubnetInfo"]),
        #
        'DhcpEnumOptionValuesV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO6", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "ClassName", "VendorName", "ScopeInfo", "ResumeHandle", "PreferredMaximum", "OptionValues", "OptionsRead", "OptionsTotal"]),
        #
        'DhcpDsInit': SimTypeFunction([], SimTypeInt(signed=False, label="UInt32")),
        #
        'DhcpDsCleanup': SimTypeFunction([], SimTypeBottom(label="Void")),
        #
        'DhcpSetThreadOptions': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Flags", "Reserved"]),
        #
        'DhcpGetThreadOptions': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pFlags", "Reserved"]),
        #
        'DhcpServerQueryAttribute': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_ATTRIB", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddr", "dwReserved", "DhcpAttribId", "pDhcpAttrib"]),
        #
        'DhcpServerQueryAttributes': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_ATTRIB_ARRAY", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddr", "dwReserved", "dwAttribCount", "pDhcpAttribs", "pDhcpAttribArr"]),
        #
        'DhcpServerRedoAuthorization': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddr", "dwReserved"]),
        #
        'DhcpAuditLogSetParams': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "AuditLogDir", "DiskCheckInterval", "MaxLogFilesSize", "MinSpaceOnDisk"]),
        #
        'DhcpAuditLogGetParams': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "AuditLogDir", "DiskCheckInterval", "MaxLogFilesSize", "MinSpaceOnDisk"]),
        #
        'DhcpServerQueryDnsRegCredentials': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "UnameSize", "Uname", "DomainSize", "Domain"]),
        #
        'DhcpServerSetDnsRegCredentials': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Uname", "Domain", "Passwd"]),
        #
        'DhcpServerSetDnsRegCredentialsV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Uname", "Domain", "Passwd"]),
        #
        'DhcpServerBackupDatabase': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Path"]),
        #
        'DhcpServerRestoreDatabase': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Path"]),
        #
        'DhcpServerSetConfigVQ': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SERVER_CONFIG_INFO_VQ", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "FieldsToSet", "ConfigInfo"]),
        #
        'DhcpServerGetConfigVQ': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SERVER_CONFIG_INFO_VQ", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ConfigInfo"]),
        #
        'DhcpGetServerSpecificStrings': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SERVER_SPECIFIC_STRINGS", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ServerSpecificStrings"]),
        #
        'DhcpServerAuditlogParamsFree': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_SERVER_CONFIG_INFO_VQ", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["ConfigInfo"]),
        #
        'DhcpCreateSubnetV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypePointer(SimTypeRef("DHCP_SUBNET_INFO_V6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SubnetInfo"]),
        #
        'DhcpDeleteSubnetV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypeInt(signed=False, label="DHCP_FORCE_FLAG")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ForceFlag"]),
        #
        'DhcpEnumSubnetsV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCPV6_IP_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ResumeHandle", "PreferredMaximum", "EnumInfo", "ElementsRead", "ElementsTotal"]),
        #
        'DhcpAddSubnetElementV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_DATA_V6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "AddElementInfo"]),
        #
        'DhcpRemoveSubnetElementV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_DATA_V6", SimStruct), offset=0), SimTypeInt(signed=False, label="DHCP_FORCE_FLAG")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "RemoveElementInfo", "ForceFlag"]),
        #
        'DhcpEnumSubnetElementsV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypeInt(signed=False, label="DHCP_SUBNET_ELEMENT_TYPE_V6"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SUBNET_ELEMENT_INFO_ARRAY_V6", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "EnumElementType", "ResumeHandle", "PreferredMaximum", "EnumElementInfo", "ElementsRead", "ElementsTotal"]),
        #
        'DhcpGetSubnetInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SUBNET_INFO_V6", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SubnetInfo"]),
        #
        'DhcpEnumSubnetClientsV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypePointer(SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_ARRAY_V6", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ResumeHandle", "PreferredMaximum", "ClientInfo", "ClientsRead", "ClientsTotal"]),
        #
        'DhcpServerGetConfigV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO6", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_SERVER_CONFIG_INFO_V6", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ScopeInfo", "ConfigInfo"]),
        #
        'DhcpServerSetConfigV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO6", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_SERVER_CONFIG_INFO_V6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ScopeInfo", "FieldsToSet", "ConfigInfo"]),
        #
        'DhcpSetSubnetInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypePointer(SimTypeRef("DHCP_SUBNET_INFO_V6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "SubnetInfo"]),
        #
        'DhcpGetMibInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_MIB_INFO_V6", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "MibInfo"]),
        #
        'DhcpGetServerBindingInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCPV6_BIND_ELEMENT_ARRAY", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "BindElementsInfo"]),
        #
        'DhcpSetServerBindingInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCPV6_BIND_ELEMENT_ARRAY", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "BindElementInfo"]),
        #
        'DhcpSetClientInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_V6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpGetClientInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_SEARCH_INFO_V6", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_V6", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SearchInfo", "ClientInfo"]),
        #
        'DhcpDeleteClientInfoV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_SEARCH_INFO_V6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpCreateClassV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_CLASS_INFO_V6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ReservedMustBeZero", "ClassInfo"]),
        #
        'DhcpModifyClassV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_CLASS_INFO_V6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ReservedMustBeZero", "ClassInfo"]),
        #
        'DhcpDeleteClassV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ReservedMustBeZero", "ClassName"]),
        #
        'DhcpEnumClassesV6': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLASS_INFO_ARRAY_V6", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ReservedMustBeZero", "ResumeHandle", "PreferredMaximum", "ClassInfoArray", "nRead", "nTotal"]),
        #
        'DhcpSetSubnetDelayOffer': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeShort(signed=False, label="UInt16")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "TimeDelayInMilliseconds"]),
        #
        'DhcpGetSubnetDelayOffer': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "TimeDelayInMilliseconds"]),
        #
        'DhcpGetMibInfoV5': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_MIB_INFO_V5", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "MibInfo"]),
        #
        'DhcpAddSecurityGroup': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pServer"]),
        #
        'DhcpV4GetOptionValue': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "PolicyName", "VendorName", "ScopeInfo", "OptionValue"]),
        #
        'DhcpV4SetOptionValue': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_DATA", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionId", "PolicyName", "VendorName", "ScopeInfo", "OptionValue"]),
        #
        'DhcpV4SetOptionValues': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_VALUE_ARRAY", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "PolicyName", "VendorName", "ScopeInfo", "OptionValues"]),
        #
        'DhcpV4RemoveOptionValue': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "OptionID", "PolicyName", "VendorName", "ScopeInfo"]),
        #
        'DhcpV4GetAllOptionValues': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_OPTION_SCOPE_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_ALL_OPTION_VALUES_PB", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "ScopeInfo", "Values"]),
        #
        'DhcpV4FailoverCreateRelationship': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_FAILOVER_RELATIONSHIP", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "pRelationship"]),
        #
        'DhcpV4FailoverSetRelationship': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCP_FAILOVER_RELATIONSHIP", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "Flags", "pRelationship"]),
        #
        'DhcpV4FailoverDeleteRelationship': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "pRelationshipName"]),
        #
        'DhcpV4FailoverGetRelationship': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_FAILOVER_RELATIONSHIP", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "pRelationshipName", "pRelationship"]),
        #
        'DhcpV4FailoverEnumRelationship': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_FAILOVER_RELATIONSHIP_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ResumeHandle", "PreferredMaximum", "pRelationship", "RelationshipRead", "RelationshipTotal"]),
        #
        'DhcpV4FailoverAddScopeToRelationship': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_FAILOVER_RELATIONSHIP", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "pRelationship"]),
        #
        'DhcpV4FailoverDeleteScopeFromRelationship': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_FAILOVER_RELATIONSHIP", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "pRelationship"]),
        #
        'DhcpV4FailoverGetScopeRelationship': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_FAILOVER_RELATIONSHIP", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ScopeId", "pRelationship"]),
        #
        'DhcpV4FailoverGetScopeStatistics': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_FAILOVER_STATISTICS", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ScopeId", "pStats"]),
        #
        'DhcpV4FailoverGetClientInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_SEARCH_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCPV4_FAILOVER_CLIENT_INFO", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SearchInfo", "ClientInfo"]),
        #
        'DhcpV4FailoverGetSystemTime': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "pTime", "pMaxAllowedDeltaTime"]),
        #
        'DhcpV4FailoverGetAddressStatus': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "pStatus"]),
        #
        'DhcpV4FailoverTriggerAddrAllocation': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "pFailRelName"]),
        #
        'DhcpHlprCreateV4Policy': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DHCP_POL_LOGIC_OPER"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["PolicyName", "fGlobalPolicy", "Subnet", "ProcessingOrder", "RootOperator", "Description", "Enabled", "Policy"]),
        #
        'DhcpHlprCreateV4PolicyEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DHCP_POL_LOGIC_OPER"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_POLICY_EX", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["PolicyName", "fGlobalPolicy", "Subnet", "ProcessingOrder", "RootOperator", "Description", "Enabled", "Policy"]),
        #
        'DhcpHlprAddV4PolicyExpr': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DHCP_POL_LOGIC_OPER"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Policy", "ParentExpr", "Operator", "ExprIndex"]),
        #
        'DhcpHlprAddV4PolicyCondition': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DHCP_POL_ATTR_TYPE"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="DHCP_POL_COMPARATOR"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Policy", "ParentExpr", "Type", "OptionID", "SubOptionID", "VendorName", "Operator", "Value", "ValueLength", "ConditionIndex"]),
        #
        'DhcpHlprAddV4PolicyRange': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0), SimTypePointer(SimTypeRef("DHCP_IP_RANGE", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Policy", "Range"]),
        #
        'DhcpHlprResetV4PolicyExpr': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Policy"]),
        #
        'DhcpHlprModifyV4PolicyExpr': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0), SimTypeInt(signed=False, label="DHCP_POL_LOGIC_OPER")], SimTypeInt(signed=False, label="UInt32"), arg_names=["Policy", "Operator"]),
        #
        'DhcpHlprFreeV4Policy': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["Policy"]),
        #
        'DhcpHlprFreeV4PolicyArray': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY_ARRAY", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["PolicyArray"]),
        #
        'DhcpHlprFreeV4PolicyEx': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY_EX", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["PolicyEx"]),
        #
        'DhcpHlprFreeV4PolicyExArray': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY_EX_ARRAY", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["PolicyExArray"]),
        #
        'DhcpHlprFreeV4DhcpProperty': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_PROPERTY", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["Property"]),
        #
        'DhcpHlprFreeV4DhcpPropertyArray': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_PROPERTY_ARRAY", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["PropertyArray"]),
        #
        'DhcpHlprFindV4DhcpProperty': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_PROPERTY_ARRAY", SimStruct), offset=0), SimTypeInt(signed=False, label="DHCP_PROPERTY_ID"), SimTypeInt(signed=False, label="DHCP_PROPERTY_TYPE")], SimTypePointer(SimTypeRef("DHCP_PROPERTY", SimStruct), offset=0), arg_names=["PropertyArray", "ID", "Type"]),
        #
        'DhcpHlprIsV4PolicySingleUC': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Policy"]),
        #
        'DhcpV4QueryPolicyEnforcement': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "fGlobalPolicy", "SubnetAddress", "Enabled"]),
        #
        'DhcpV4SetPolicyEnforcement': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "fGlobalPolicy", "SubnetAddress", "Enable"]),
        #
        'DhcpHlprIsV4PolicyWellFormed': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pPolicy"]),
        #
        'DhcpHlprIsV4PolicyValid': SimTypeFunction([SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pPolicy"]),
        #
        'DhcpV4CreatePolicy': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "pPolicy"]),
        #
        'DhcpV4GetPolicy': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "fGlobalPolicy", "SubnetAddress", "PolicyName", "Policy"]),
        #
        'DhcpV4SetPolicy': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_POLICY", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "FieldsModified", "fGlobalPolicy", "SubnetAddress", "PolicyName", "Policy"]),
        #
        'DhcpV4DeletePolicy': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "fGlobalPolicy", "SubnetAddress", "PolicyName"]),
        #
        'DhcpV4EnumPolicies': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_POLICY_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ResumeHandle", "PreferredMaximum", "fGlobalPolicy", "SubnetAddress", "EnumInfo", "ElementsRead", "ElementsTotal"]),
        #
        'DhcpV4AddPolicyRange': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_IP_RANGE", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "PolicyName", "Range"]),
        #
        'DhcpV4RemovePolicyRange': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_IP_RANGE", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "PolicyName", "Range"]),
        #
        'DhcpV6SetStatelessStoreParams': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DHCPV6_STATELESS_PARAMS", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "fServerLevel", "SubnetAddress", "FieldModified", "Params"]),
        #
        'DhcpV6GetStatelessStoreParams': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypePointer(SimTypePointer(SimTypeRef("DHCPV6_STATELESS_PARAMS", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "fServerLevel", "SubnetAddress", "Params"]),
        #
        'DhcpV6GetStatelessStatistics': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCPV6_STATELESS_STATS", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "StatelessStats"]),
        #
        'DhcpV4CreateClientInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_PB", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpV4EnumSubnetClients': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_PB_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ResumeHandle", "PreferredMaximum", "ClientInfo", "ClientsRead", "ClientsTotal"]),
        #
        'DhcpV4GetClientInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_SEARCH_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_PB", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SearchInfo", "ClientInfo"]),
        #
        'DhcpV6CreateClientInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_V6", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpV4GetFreeIPAddress': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_IP_ARRAY", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ScopeId", "StartIP", "EndIP", "NumFreeAddrReq", "IPAddrList"]),
        #
        'DhcpV6GetFreeIPAddress': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypeRef("DHCP_IPV6_ADDRESS", SimStruct), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCPV6_IP_ARRAY", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ScopeId", "StartIP", "EndIP", "NumFreeAddrReq", "IPAddrList"]),
        #
        'DhcpV4CreateClientInfoEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_EX", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ClientInfo"]),
        #
        'DhcpV4EnumSubnetClientsEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_EX_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SubnetAddress", "ResumeHandle", "PreferredMaximum", "ClientInfo", "ClientsRead", "ClientsTotal"]),
        #
        'DhcpV4GetClientInfoEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_SEARCH_INFO", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_CLIENT_INFO_EX", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "SearchInfo", "ClientInfo"]),
        #
        'DhcpV4CreatePolicyEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_POLICY_EX", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "PolicyEx"]),
        #
        'DhcpV4GetPolicyEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_POLICY_EX", SimStruct), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "GlobalPolicy", "SubnetAddress", "PolicyName", "Policy"]),
        #
        'DhcpV4SetPolicyEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("DHCP_POLICY_EX", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "FieldsModified", "GlobalPolicy", "SubnetAddress", "PolicyName", "Policy"]),
        #
        'DhcpV4EnumPoliciesEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("DHCP_POLICY_EX_ARRAY", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ServerIpAddress", "ResumeHandle", "PreferredMaximum", "GlobalPolicy", "SubnetAddress", "EnumInfo", "ElementsRead", "ElementsTotal"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("rtutils.dll")
prototypes = \
    {
        #
        'TraceRegisterExA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpszCallerName", "dwFlags"]),
        #
        'TraceDeregisterA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID"]),
        #
        'TraceDeregisterExA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags"]),
        #
        'TraceGetConsoleA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "lphConsole"]),
        #
        'TracePrintfA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "lpszFormat"]),
        #
        'TracePrintfExA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags", "lpszFormat"]),
        #
        'TraceVprintfExA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags", "lpszFormat", "arglist"]),
        #
        'TracePutsExA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags", "lpszString"]),
        #
        'TraceDumpExA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags", "lpbBytes", "dwByteCount", "dwGroupSize", "bAddressPrefix", "lpszPrefix"]),
        #
        'TraceRegisterExW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpszCallerName", "dwFlags"]),
        #
        'TraceDeregisterW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID"]),
        #
        'TraceDeregisterExW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags"]),
        #
        'TraceGetConsoleW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "lphConsole"]),
        #
        'TracePrintfW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "lpszFormat"]),
        #
        'TracePrintfExW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags", "lpszFormat"]),
        #
        'TraceVprintfExW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags", "lpszFormat", "arglist"]),
        #
        'TracePutsExW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags", "lpszString"]),
        #
        'TraceDumpExW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTraceID", "dwFlags", "lpbBytes", "dwByteCount", "dwGroupSize", "bAddressPrefix", "lpszPrefix"]),
        #
        'LogErrorA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["dwMessageId", "cNumberOfSubStrings", "plpwsSubStrings", "dwErrorCode"]),
        #
        'LogEventA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), label="LPArray", offset=0)], SimTypeBottom(label="Void"), arg_names=["wEventType", "dwMessageId", "cNumberOfSubStrings", "plpwsSubStrings"]),
        #
        'LogErrorW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["dwMessageId", "cNumberOfSubStrings", "plpwsSubStrings", "dwErrorCode"]),
        #
        'LogEventW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0)], SimTypeBottom(label="Void"), arg_names=["wEventType", "dwMessageId", "cNumberOfSubStrings", "plpwsSubStrings"]),
        #
        'RouterLogRegisterA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["lpszSource"]),
        #
        'RouterLogDeregisterA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["hLogHandle"]),
        #
        'RouterLogEventA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwMessageId", "dwSubStringCount", "plpszSubStringArray", "dwErrorCode"]),
        #
        'RouterLogEventDataA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwMessageId", "dwSubStringCount", "plpszSubStringArray", "dwDataBytes", "lpDataBytes"]),
        #
        'RouterLogEventStringA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwMessageId", "dwSubStringCount", "plpszSubStringArray", "dwErrorCode", "dwErrorIndex"]),
        #
        'RouterLogEventExA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwErrorCode", "dwMessageId", "ptszFormat"]),
        #
        'RouterLogEventValistExA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwErrorCode", "dwMessageId", "ptszFormat", "arglist"]),
        #
        'RouterGetErrorStringA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwErrorCode", "lplpszErrorString"]),
        #
        'RouterLogRegisterW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["lpszSource"]),
        #
        'RouterLogDeregisterW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["hLogHandle"]),
        #
        'RouterLogEventW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwMessageId", "dwSubStringCount", "plpszSubStringArray", "dwErrorCode"]),
        #
        'RouterLogEventDataW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwMessageId", "dwSubStringCount", "plpszSubStringArray", "dwDataBytes", "lpDataBytes"]),
        #
        'RouterLogEventStringW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwMessageId", "dwSubStringCount", "plpszSubStringArray", "dwErrorCode", "dwErrorIndex"]),
        #
        'RouterLogEventExW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwErrorCode", "dwMessageId", "ptszFormat"]),
        #
        'RouterLogEventValistExW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["hLogHandle", "dwEventType", "dwErrorCode", "dwMessageId", "ptszFormat", "arglist"]),
        #
        'RouterGetErrorStringW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwErrorCode", "lplpwszErrorString"]),
        #
        'RouterAssert': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["pszFailedAssertion", "pszFileName", "dwLineNumber", "pszMessage"]),
        #
        'MprSetupProtocolEnum': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwTransportId", "lplpBuffer", "lpdwEntriesRead"]),
        #
        'MprSetupProtocolFree': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpBuffer"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations
from .base import SimSootValue


class SimSootValue_IntConstant(SimSootValue):
    __slots__ = ["type", "value"]

    def __init__(self, value, type_):
        super().__init__()
        self.value = value
        self.type = type_

    def __repr__(self):
        return self.value

    @classmethod
    def from_sootvalue(cls, soot_value, state):
        return cls(soot_value.value, soot_value.type)



from __future__ import annotations
from typing import Any
from collections.abc import Callable
from functools import partial
import logging

import networkx

import angr.ailment as ailment
from angr.ailment import Block
from angr.ailment.expression import Phi, VirtualVariable, VirtualVariableCategory
from angr.ailment.statement import Assignment, Label

from angr.code_location import CodeLocation
from angr.analyses import ForwardAnalysis
from angr.analyses.forward_analysis import FunctionGraphVisitor
from angr.utils.ail import is_head_controlled_loop_block
from .rewriting_engine import SimEngineSSARewriting, DefExprType, AT
from .rewriting_state import RewritingState

l = logging.getLogger(__name__)


class RewritingAnalysis(ForwardAnalysis[RewritingState, ailment.Block, object, object]):
    """
    RewritingAnalysis traverses the AIL graph, inserts phi nodes, and rewrites all expression uses to virtual variables
    when necessary.
    """

    def __init__(
        self,
        project,
        func,
        ail_graph,
        sp_tracker,
        bp_as_gpr: bool,
        udef_to_phiid: dict[tuple, set[int]],
        phiid_to_loc: dict[int, tuple[int, int | None]],
        stackvar_locs: dict[int, set[int]],
        rewrite_tmps: bool,
        ail_manager,
        func_args: set[VirtualVariable],
        vvar_id_start: int = 0,
    ):
        self.project = project
        self._function = func
        self._graph_visitor = FunctionGraphVisitor(self._function, ail_graph)

        ForwardAnalysis.__init__(
            self, order_jobs=True, allow_merging=False, allow_widening=False, graph_visitor=self._graph_visitor
        )
        self._graph = ail_graph
        self._udef_to_phiid = udef_to_phiid
        self._phiid_to_loc = phiid_to_loc
        self._stackvar_locs = stackvar_locs
        self._rewrite_tmps = rewrite_tmps
        self._ail_manager = ail_manager
        self._func_args = func_args
        self._engine_ail = SimEngineSSARewriting(
            self.project,
            sp_tracker=sp_tracker,
            bp_as_gpr=bp_as_gpr,
            udef_to_phiid=self._udef_to_phiid,
            phiid_to_loc=self._phiid_to_loc,
            stackvar_locs=self._stackvar_locs,
            rewrite_tmps=self._rewrite_tmps,
            ail_manager=ail_manager,
            vvar_id_start=vvar_id_start,
        )

        self._visited_blocks: set[Any] = set()
        self.out_blocks = {}
        self.out_states = {}
        # loop_states stores states at the beginning of a loop block *after a loop iteration*, where the block is the
        # following:
        #    0x4036df | t4 = (rcx<8> == 0x0<64>)
        #    0x4036df | if (t4) { Goto 0x4036e2<64> } else { Goto 0x4036df<64> }
        #    0x4036df | STORE(addr=t3, data=t2, size=8, endness=Iend_LE, guard=None)
        #    0x4036df | rdi<8> = t8
        #
        self.head_controlled_loop_outstates = {}

        self._analyze()

        self.def_to_vvid: dict[tuple[int, int | None, int, DefExprType, AT], int] = self._engine_ail.def_to_vvid
        # during SSA conversion, we create secondary stack variables because they overlap and are larger than the
        # actual stack variables. these secondary stack variables can be safely eliminated during dead assignment
        # elimination if not used by anything else.
        self.secondary_stackvars: set[int] = self._engine_ail.secondary_stackvars
        self.out_graph = self._make_new_graph(ail_graph)

    @property
    def max_vvar_id(self) -> int | None:
        return self._engine_ail.current_vvar_id

    def _make_new_graph(self, old_graph: networkx.DiGraph) -> networkx.DiGraph:
        new_graph = networkx.DiGraph()
        for node in old_graph:
            new_graph.add_node(self.out_blocks.get((node.addr, node.idx), node))

        for src, dst, data in old_graph.edges(data=True):
            new_graph.add_edge(
                self.out_blocks.get((src.addr, src.idx), src),
                self.out_blocks.get((dst.addr, dst.idx), dst),
                **data,
            )

        return new_graph

    def create_phi_statements(self, node, udef_to_phiid: dict, phiid_to_loc: dict) -> list[Assignment]:
        phi_ids = []
        for phiid, (block_addr, block_idx) in phiid_to_loc.items():
            if block_addr == node.addr and block_idx == node.idx:
                phi_ids.append(phiid)

        phiid_to_udef = {}
        for udef, phiids in udef_to_phiid.items():
            for phiid in phiids:
                phiid_to_udef[phiid] = udef

        phi_stmts = []
        for phi_id in phi_ids:
            udef = phiid_to_udef[phi_id]
            category = udef[0]

            match category:
                case "reg":
                    _, reg_offset, reg_bits = udef

                    phi_var = Phi(
                        self._ail_manager.next_atom(),
                        reg_bits,
                        src_and_vvars=[],  # back patch later
                        ins_addr=node.addr,
                    )
                    phi_dst = VirtualVariable(
                        self._ail_manager.next_atom(),
                        self._engine_ail.next_vvar_id(),
                        reg_bits,
                        VirtualVariableCategory.REGISTER,
                        oident=reg_offset,
                        ins_addr=node.addr,
                    )

                case "stack":
                    _, stack_offset, stack_size = udef

                    phi_var = Phi(
                        self._ail_manager.next_atom(),
                        stack_size * self.project.arch.byte_width,
                        src_and_vvars=[],  # back patch later
                        ins_addr=node.addr,
                    )
                    phi_dst = VirtualVariable(
                        self._ail_manager.next_atom(),
                        self._engine_ail.next_vvar_id(),
                        stack_size * self.project.arch.byte_width,
                        VirtualVariableCategory.STACK,
                        oident=stack_offset,
                        ins_addr=node.addr,
                    )
                case _:
                    raise NotImplementedError

            phi_stmt = Assignment(
                None,
                phi_dst,
                phi_var,
                ins_addr=node.addr,
            )
            phi_stmts.append(phi_stmt)

        return phi_stmts

    @staticmethod
    def insert_phi_statements(node: Block, phi_stmts: list[Assignment]):
        idx = 0
        while idx < len(node.statements):
            if not isinstance(node.statements[idx], Label):
                break
            idx += 1

        if idx >= len(node.statements):
            node.statements += phi_stmts
        else:
            node.statements = node.statements[:idx] + phi_stmts + node.statements[idx:]

    def _reg_predicate(self, node_: Block, *, reg_offset: int, reg_size: int) -> tuple[bool, Any]:
        out_state: RewritingState = (
            self.head_controlled_loop_outstates[(node_.addr, node_.idx)]
            if is_head_controlled_loop_block(node_)
            else self.out_states[(node_.addr, node_.idx)]
        )
        if reg_offset in out_state.registers and reg_size in out_state.registers[reg_offset]:
            existing_var = out_state.registers[reg_offset][reg_size]
            if existing_var is None:
                # the vvar is not set. it should never be referenced
                return True, None
            vvar = existing_var.copy()
            vvar.idx = self._ail_manager.next_atom()
            return True, vvar
        return False, None

    def _stack_predicate(self, node_: Block, *, stack_offset: int, stackvar_size: int) -> tuple[bool, Any]:
        out_state: RewritingState = (
            self.head_controlled_loop_outstates[(node_.addr, node_.idx)]
            if is_head_controlled_loop_block(node_)
            else self.out_states[(node_.addr, node_.idx)]
        )
        if stack_offset in out_state.stackvars and stackvar_size in out_state.stackvars[stack_offset]:
            existing_var = out_state.stackvars[stack_offset][stackvar_size]
            if existing_var is None:
                # the vvar is not set. it should never be referenced
                return True, None
            vvar = existing_var.copy()
            vvar.idx = self._ail_manager.next_atom()
            return True, vvar
        return False, None

    #
    # Main analysis routines
    #

    def _pre_analysis(self):
        for node in self.graph:
            phi_stmts = self.create_phi_statements(node, self._udef_to_phiid, self._phiid_to_loc)
            if phi_stmts:
                self.insert_phi_statements(node, phi_stmts)

    def _initial_abstract_state(self, node) -> RewritingState:
        state = RewritingState(
            CodeLocation(node.addr, stmt_idx=0, ins_addr=node.addr, block_idx=node.idx),
            self.project.arch,
            self._function,
            node,
        )
        # update state with function arguments
        for func_arg in self._func_args:
            if func_arg.parameter_category == VirtualVariableCategory.REGISTER:
                reg_offset, reg_size = func_arg.parameter_reg_offset, func_arg.size
                assert reg_offset is not None and reg_size is not None
                state.registers[reg_offset][reg_size] = func_arg
            elif func_arg.parameter_category == VirtualVariableCategory.STACK:
                parameter_stack_offset: int = func_arg.oident[1]  # type: ignore
                assert parameter_stack_offset is not None and func_arg.size is not None
                state.stackvars[parameter_stack_offset][func_arg.size] = func_arg
        return state

    def _run_on_node(self, node, state: RewritingState):
        """

        :param node:    The current node.
        :param state:   The analysis state.
        :return:        A tuple: (any changes occur, successor state)
        """

        if isinstance(node, ailment.Block):
            block = node
            block_key = (node.addr, node.idx)
            engine = self._engine_ail
        else:
            l.warning("Unsupported node type %s.", node.__class__)
            return False, state

        if block_key in self._visited_blocks:
            return False, state

        engine: SimEngineSSARewriting

        old_state = state
        state = old_state.copy()
        state.loc = CodeLocation(block.addr, stmt_idx=0, ins_addr=block.addr, block_idx=block.idx)
        state.original_block = node

        engine.process(
            state,
            block=block,
        )

        self._visited_blocks.add(block_key)
        # get the output state (which is the input state for the successor node)
        # if head_controlled_loop_outstate is set, then it is the output state of the successor node; in this case, the
        # input state for the head-controlled loop block itself is out.state.
        # otherwise (if head_controlled_loop_outstate is not set), engine.state is the input state of the successor
        # node.
        if engine.head_controlled_loop_outstate is None:
            # this is a normal block
            out_state = state
        else:
            # this is a head-controlled loop block
            out_state = engine.head_controlled_loop_outstate
            self.head_controlled_loop_outstates[block_key] = state
        self.out_states[block_key] = out_state
        # the final block is always in state
        out_block = state.out_block

        if out_block is not None:
            assert out_block.addr == block.addr

            if self.out_blocks.get(block_key, None) == out_block:
                return True, out_state
            self.out_blocks[block_key] = out_block
            out_state.out_block = None
            return True, out_state

        return True, out_state

    def _intra_analysis(self):
        pass

    def _post_analysis(self):
        # update phi nodes
        for original_node in self.graph:
            node_key = original_node.addr, original_node.idx
            node = self.out_blocks.get(node_key, original_node)

            if any(
                isinstance(stmt, Assignment) and isinstance(stmt.dst, VirtualVariable) and isinstance(stmt.src, Phi)
                for stmt in node.statements
            ):
                # there we go
                new_stmts = []
                for stmt in node.statements:
                    if not (
                        isinstance(stmt, Assignment)
                        and isinstance(stmt.dst, VirtualVariable)
                        and isinstance(stmt.src, Phi)
                        and len(stmt.src.src_and_vvars) == 0  # avoid re-assignment
                    ):
                        new_stmts.append(stmt)
                        continue

                    src_and_vvars = []
                    if stmt.dst.was_reg:
                        for pred in self.graph.predecessors(original_node):
                            vvar = self._follow_one_path_backward(
                                self.graph,
                                pred,
                                partial(self._reg_predicate, reg_offset=stmt.dst.reg_offset, reg_size=stmt.dst.size),
                            )
                            src_and_vvars.append(((pred.addr, pred.idx), vvar))
                    elif stmt.dst.was_stack:
                        for pred in self.graph.predecessors(original_node):
                            vvar = self._follow_one_path_backward(
                                self.graph,
                                pred,
                                partial(
                                    self._stack_predicate,
                                    stack_offset=stmt.dst.stack_offset,
                                    stackvar_size=stmt.dst.size,
                                ),
                            )
                            src_and_vvars.append(((pred.addr, pred.idx), vvar))
                    else:
                        raise NotImplementedError

                    phi_var = Phi(stmt.src.idx, stmt.src.bits, src_and_vvars=src_and_vvars)
                    new_stmt = Assignment(stmt.idx, stmt.dst, phi_var, **stmt.tags)
                    new_stmts.append(new_stmt)
                node = node.copy(statements=new_stmts)
                self.out_blocks[node_key] = node

    @staticmethod
    def _follow_one_path_backward(graph: networkx.DiGraph, src, predicate: Callable) -> Any:
        visited = set()
        return_value = None
        the_node = src
        while the_node not in visited:
            visited.add(the_node)
            stop, return_value = predicate(the_node)
            if stop:
                break
            # keep going
            more_preds = list(graph.predecessors(the_node))
            if len(more_preds) != 1:
                # no longer a single path back
                break
            the_node = more_preds[0]
        return return_value
